//     mpibart.cpp
//     Copyright (C) 2012-2016 Matthew T. Pratola and Robert E. McCulloch
//
//     This file is part of MPIBART.
//
//     MPIBART is free software: you can redistribute it and/or modify
//     it under the terms of the GNU Affero General Public License as published by
//     the Free Software Foundation, either version 3 of the License, or
//     (at your option) any later version.
//
//     MPIBART is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU Affero General Public License for more details.
//
//     You should have received a copy of the GNU Affero General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//     Author contact information
//     Matthew T. Pratola: mpratola@gmail.com

#include <iostream>
#include <fstream>
#include <vector>
#include <ctime>
#include <sstream>

#include "rng.h"
#include "tree.h"
#include "info.h"
#include "funs.h"
#include "bd.h"
#include "mpi.h"

using std::cout;
using std::endl;

void xitofile(std::ofstream& fnm, xinfo xi)
{
   fnm << xi.size() << endl;
   for(uint i=0;i< xi.size();i++) {
      fnm << xi[i].size() << endl;
      for(uint j=0;j<xi[i].size();j++) fnm << xi[i][j] << "  ";
      fnm << endl;
   }
   fnm << endl;
}

int main(int argc, char** argv)
{
	int myrank, processes;
	MPI_Init(&argc,&argv);
	MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
	MPI_Comm_size(MPI_COMM_WORLD,&processes);

#ifndef SILENT
	cout << "\nMPI: node " << myrank << " of " << processes << " processes.";
   cout << "\n*****Into bart main\n";
#endif
   if(argc<3) {cout << "error need at least 2 arguments\n"; return 1;}
   std::stringstream xroot(argv[1]);  // this is the root of the x and y filenames for bart to read data
   std::stringstream yroot(argv[2]);  // in from.  The format of data files must be <root>#.txt. The number
                                  // represented by # is the node number for the parallel slave's, and 
                                  // as such # must be greater than 0.
   double xtemp,ytemp; //used to read in x and y without knowing how many there are.

   //--------------------------------------------------
   //random number generation
   uint seed=99;
	MPI_Bcast(&seed,1,MPI_UNSIGNED,0,MPI_COMM_WORLD);
   RNG gen(seed); //this one random number generator is used in all draws
#ifndef SILENT
	cout << "RNG seed set to " << seed << "\n";
#endif

   //--------------------------------------------------
   //read in data
   //read y NOTE: we assume y is already centered at 0.
	//IMPORTANT NOTES FOR MPI VERSION: 
	// allys has correct information on the root node only
	// miny and maxy have the min/max of the y's loaded on the each slave respectively, and the min/max of the whole
	// dataset on the root node
	// n is the size of the entire dataset, nn is the size of the dataset loaded on each slave respectively
	sinfo allys;
	double sy=0.0, sy2=0.0;
	unsigned int tempn=0;
	size_t n=0,nn=0;
	double miny=INFINITY;
	double maxy=-INFINITY;
	double myminy=INFINITY;
	double mymaxy=-INFINITY;
	std::vector<double> y;   //storage for y on the slaves
	if(myrank!=0)
	{
		std::stringstream yfss;
		std::string yfs;
//		yfss << "y" << myrank << ".txt";
		yfss << yroot.str() << myrank << ".txt";
		yfs=yfss.str();
#ifndef SILENT
		cout << myrank << ": loading y's from " << yfs << endl;
#endif
		std::ifstream yf(yfs.c_str()); //file to read y from for this slave
	
		while(yf >> ytemp) {
			y.push_back(ytemp);
			if(ytemp<myminy) myminy=ytemp;
			if(ytemp>mymaxy) mymaxy=ytemp;
			sy+=ytemp;
			sy2+=ytemp*ytemp;
		}
		tempn=y.size();
		nn=(size_t)tempn;
		if(nn<1) {
			cout << "error nn<1\n";
			MPI_Finalize();
			return 1;
		}
#ifndef SILENT
		cout << myrank << ": nn=" << nn << " sy=" << sy << " sy2=" << sy2 << " myminy=" << myminy << " mymaxy=" << mymaxy << endl;
#endif
		miny=myminy;
		maxy=mymaxy;
	}
	MPI_Allreduce(&sy,&allys.sy,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
	MPI_Allreduce(&sy2,&allys.sy2,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
	MPI_Allreduce(&tempn,&tempn,1,MPI_UNSIGNED,MPI_SUM,MPI_COMM_WORLD);
	MPI_Allreduce(&myminy,&miny,1,MPI_DOUBLE,MPI_MIN,MPI_COMM_WORLD);
	MPI_Allreduce(&mymaxy,&maxy,1,MPI_DOUBLE,MPI_MAX,MPI_COMM_WORLD);
	n=(size_t) tempn;
#ifndef SILENT
	cout << myrank << ": n=" << n << " nn=" << nn << " allys.sy=" << allys.sy << " allys.sy2=" << allys.sy2 << " miny=" << miny << " maxy=" << maxy << "\n";
#endif
	if(n<1) {
		cout << "error n<1\n";
		MPI_Finalize();
		return 1;
	}
	allys.n=n;
   double ybar = allys.sy/n; //sample mean
   double shat = sqrt((allys.sy2-n*ybar*ybar)/(n-1)); //sample standard deviation
#ifndef SILENT
   cout << "ybar,shat: " << ybar << ", " << shat <<  endl;
#endif

   //read x
   //the n*p numbers for x are stored as the p for first obs, then p for second, and so on.
	unsigned int myp=0, tempp=0;
	size_t p=0;
	std::vector<double> x;
	if(myrank!=0)
	{
		std::stringstream xfss;
		std::string xfs;
//		xfss << "x" << myrank << ".txt";
		xfss << xroot.str() << myrank << ".txt";
		xfs=xfss.str();
#ifndef SILENT
		cout << myrank << ": loading x's from " << xfs << endl;
#endif
		std::ifstream xf(xfs.c_str());  //file to read x from
   	while(xf >> xtemp) 
   	   x.push_back(xtemp);
   	myp = (unsigned int)x.size()/nn;
		p = (size_t) myp;
		if(x.size() != nn*p) {
      	cout << "error: input x file has wrong number of values\n"; 
			MPI_Finalize();
      	return 1;
		}
	}
	MPI_Reduce(&myp,&tempp,1,MPI_UNSIGNED,MPI_SUM,0,MPI_COMM_WORLD);
	if(myrank==0)
	{
		p=(size_t) tempp/(processes-1);
		if(p!=tempp/(processes-1))
		{
			cout << "error: input x files on compute nodes have different number of predictors p\n";
			MPI_Finalize();
			return 1;
		}
	}
#ifndef SILENT
	cout << myrank << ": myp=" << myp << " p=" << p << endl;
#endif

   //x for predictions
   dinfo dip; //data information for prediction
   dip.n=0;
   //Currently no out-of-sample prediction in MPIBART
#ifndef SILENT
   cout << "\nx for prediction read in:\n";
   cout << "n: " << dip.n << endl;
#endif
   if(dip.n) {
#ifndef SILENT
      cout << "first row: " <<  dip.x[0] << " ...  " << dip.x[p-1] << endl;
      cout << "last row: " << dip.x[(dip.n-1)*p] << " ...  " << dip.x[dip.n*p-1] << endl;
#endif
   }

   //--------------------------------------------------
   //optionally read in additional arguments
   size_t burn = 1000; //number of mcmc iterations called burn-in
   size_t nd = 1000; //number of mcmc iterations
   size_t thin = 10; //save a draw ever thin iterations after burn-in.
   size_t m=200; //number of trees in BART sum
   double lambda = 1.0; //this one really needs to be set
   double nu = 3.0;
   double kfac=2.0;
   std::stringstream fname("");  //root name of output files 

   if(argc>3) thin = atoi(argv[3]);
   if(argc>4) nd = atoi(argv[4]);
   if(argc>5) lambda = atof(argv[5]);
   if(argc>6) burn = atoi(argv[6]);
   if(argc>7) m = atoi(argv[7]);
   if(argc>8) nu = atof(argv[8]);
   if(argc>9) kfac = atof(argv[9]);
   if(argc>10) fname << argv[10];

#ifndef SILENT
   cout <<"\nburn,nd,number of trees: " << burn << ", " << nd << ", " << m << endl;
   cout <<"\nlambda,nu,kfac: " << lambda << ", " << nu << ", " << kfac << endl;
   cout <<"\noutput file root is: -->" << fname.str() << "<--" << endl;
#endif

   //--------------------------------------------------
   //x cutpoints
	std::vector<double> minx(p,INFINITY);
   std::vector<double> maxx(p,-INFINITY);
	xinfo xi;
	size_t nc=100; //100 equally spaced cutpoints from min to max.

	if(myrank!=0)
		makeminmax(p,nn,&x[0],minx,maxx);
	for(size_t i=0;i<p;i++) {		
		MPI_Allreduce(&minx[i],&minx[i],1,MPI_DOUBLE,MPI_MIN,MPI_COMM_WORLD);
		MPI_Allreduce(&maxx[i],&maxx[i],1,MPI_DOUBLE,MPI_MAX,MPI_COMM_WORLD);
#ifndef SILENT
		cout << myrank << ": minx[" << i << "]=" << minx[i] << " maxx[" << i << "]=" << maxx[i] << endl;
#endif
	}
	makexinfominmax(p,xi,nc,minx,maxx); //need to consruct the cutpoints on master and slaves.

   //--------------------------------------------------
   //trees
#ifndef SILENT
	cout << myrank << ": ybar=" << ybar << "\n";
#endif
   std::vector<tree> t(m);
   for(size_t i=0;i<m;i++) t[i].setm(ybar/m); //if you sum the fit over the trees you get the fit.

   //--------------------------------------------------
  //prior and mcmc
   pinfo pi;
   pi.pbd=1.0; //prob of birth/death move
   pi.pb=.5; //prob of birth given  birth/death

   pi.alpha=.95; //.95 prior prob a bot node splits is alpha/(1+d)^beta, d is depth of node
   pi.beta=2; //2 for bart means it is harder to build big trees.
   pi.tau=(maxy-miny)/(2*kfac*sqrt((double)m));
   pi.sigma=shat;

#ifndef SILENT
   cout << "\nalpha, beta: " << pi.alpha << ", " << pi.beta << endl;
   cout << "sigma, tau: " << pi.sigma << ", " << pi.tau << endl;
#endif

   //--------------------------------------------------
   //dinfo
	dinfo di;
	double *ftemp=NULL;
	double *r=NULL;
	double *allfit=NULL;
	
	if(myrank==0)
	{
		di.n=n; di.p=p; di.x=NULL; di.y=NULL;
	}
	else
	{
	   allfit = new double[nn]; //sum of fit of all trees
   	for(size_t i=0;i<nn;i++) allfit[i]=ybar;
	   r = new double[nn]; //y-(allfit-ftemp) = y-allfit+ftemp
	   ftemp = new double[nn]; //fit of current tree
	   di.n=nn; di.p=p; di.x = &x[0]; di.y=r; //the y for each draw will be the residual 		
	}

   //--------------------------------------------------
   //storage for ouput
   //in sample fit
	double *pmean;
	//for sigma draw
	double rss; //residual sum of squares
	double restemp; //a residual
  	//out of sample fit
 	double* ppredmean=0; //posterior mean for prediction
	double* fpredtemp=0; //temporary fit vector to compute prediction
  std::stringstream tempfnss;  // tempfnss and tempfns used to build the
  std::string tempfns;         // output filenames 

  tempfnss << fname.str() << "bart-sd.txt";
  tempfns = tempfnss.str();
  std::ofstream bsdf(tempfns.c_str());
    //note that we write all burn+nd draws to this file.

	if(myrank!=0)
	{
	   pmean = new double[nn]; //posterior mean of in-sample fit, sum draws,then divide
   	for(size_t i=0;i<nn;i++) pmean[i]=0.0;

   	if(dip.n) {
      	ppredmean = new double[dip.n];
      	fpredtemp = new double[dip.n];
      	for(size_t i=0;i<dip.n;i++) ppredmean[i]=0.0;
   	}
	}

   //--------------------------------------------------
   //mcmc
	double temprss;
	MPI_Request request;
	MPI_Status status;
	if(myrank==0)
	{
		//Setup saving of the fitted trees
    tempfnss.str(""); // erase previous text in tempfnss 
    tempfnss << fname.str() << "trees.txt";
    tempfns = tempfnss.str();
		std::ofstream treef(tempfns.c_str());
		//treef << xi << endl; //cutpoints
                xitofile(treef,xi);
		treef << m << endl;  //number of trees
		treef << p << endl;  //dimension of x's
		treef << (int)(nd/thin) << endl; //number of draws from posterior that will be saved

    tempfnss.str(""); // erase previous text in tempfnss 
    tempfnss << fname.str() << "time.txt";
    tempfns = tempfnss.str();
    std::ofstream timef(tempfns.c_str());

#ifndef SILENT
		cout << "\nMCMC:\n";
#endif
		clock_t tp;
		tp = clock();
		for(size_t i=0;i<(nd+burn);i++)
		{
#ifndef SILENT
			if(i%10==0) cout << "i: " << i << endl;
#endif
			//draw trees
			for(size_t j=0;j<m;j++) 
			{
				bd(t[j],xi,pi,gen,processes-1);
				MPImasterdrmu(t[j],xi,pi,gen,processes-1);
			}
			//draw sigma
			rss=0.0;
			for(int j=1;j<=(processes-1);j++)
			{
				MPI_Recv(&temprss,1,MPI_DOUBLE,MPI_ANY_SOURCE,0,MPI_COMM_WORLD,&status);
				rss+=temprss;
			}
			pi.sigma = sqrt((nu*lambda + rss)/gen.chi_square(nu+n));
         		bsdf << pi.sigma << endl;
			if(i>=burn && !(((int)i)%((int)thin)))
			{
				for(size_t j=0;j<m;j++) treef << t[j] << endl;  //save all the trees from this draw
			}
		} // end of mcmc loop

                tp=clock()-tp;
                double thetime = (double)(tp)/(double)(CLOCKS_PER_SEC);
#ifndef SILENT
		cout << "time for loop: " << thetime << endl;
#endif
                timef << thetime << endl;
 	}
	else
	{
#ifndef SILENT
		cout << "slave node " << myrank << " is up" << endl;
#endif
		for(size_t i=0;i<(nd+burn);i++)
		{
			for(size_t j=0;j<m;j++)
			{
				fit(t[j],xi,di,ftemp);
				for(size_t k=0;k<nn;k++)
				{
					allfit[k] = allfit[k]-ftemp[k];
					r[k] = y[k]-allfit[k];
				}
				// Slave code to handle the bd() call on the master
				MPIslavegetsuff(t[j],xi,di);
				MPIslaveupdatebirthdeath(t[j]);
				// Slave code to handle the drmu() call on the master
				MPIslavedrmu(t[j],xi,di);
				fit(t[j],xi,di,ftemp);
				for(size_t k=0;k<nn;k++) allfit[k] += ftemp[k];
			}
			// Slave code to handle the draw sigma code on the master
			rss=0.0;
			for(size_t k=0;k<nn;k++) {restemp=y[k]-allfit[k]; rss += restemp*restemp;}
			MPI_Isend(&rss,1,MPI_DOUBLE,0,0,MPI_COMM_WORLD,&request);

			if(i>=burn)
			{
				for(size_t k=0;k<nn;k++) pmean[k] += allfit[k];
				if(dip.n)
				{
					for(size_t j=0;j<m;j++)
					{
						fit(t[j],xi,dip,fpredtemp);
						for(size_t k=0;k<dip.n;k++) ppredmean[k] += fpredtemp[k];
					}
				}
			}
			MPI_Wait(&request,MPI_STATUS_IGNORE); //can't go further until the rss send completed.

		} //end of mcmc loop

		for(size_t k=0; k<nn; k++) pmean[k] /= (double)nd;
		std::stringstream ofss;
		std::string ofs;
		ofss << fname.str() << "bart-fit" << myrank << ".txt";
		//ofss << "bart-fit" << myrank << ".txt";
		ofs=ofss.str();
#ifndef SILENT
		cout << myrank << ": saving fit to " << ofs << endl;
#endif
		std::ofstream bfitf(ofs.c_str());
		for(size_t i=0; i<nn; i++) bfitf << pmean[i] << endl;

		//cleanup
		delete[] allfit;
		delete[] r;
		delete[] ftemp;
		delete[] pmean;
		delete[] ppredmean;
		delete[] fpredtemp;
	}

	MPI_Finalize();
	return 0;
}

